@@ -12,6 +12,7 @@ import time
 
 from .docker import DockerBuildHelper
 from .exceptions import (
+    DeploymentBranchAlreadyExisting,
     GitLabUnexpectedError,
     PaasProvisioningError,
     PaasResourceError,
@@ -19,7 +20,7 @@ from .exceptions import (
 from .job import get_job_variable
 from .runner import PaasRunner
 
-logger = logging.getLogger(__file__)
+logger = logging.getLogger(__name__)
 
 FLAVOR_JOB_VARIABLE_NAME = 'CI_CLEVER_CLOUD_FLAVOR'
 DEFAULT_FLAVOR = 'M'
@@ -60,6 +61,14 @@ class CleverCloudDockerRunner(PaasRunner):
         self.cc_zone = self.config.get('cc_zone', 'par')
         self.cc_default_flavor = self.config.get('cc_default_flavor',
                                                  DEFAULT_FLAVOR)
+        self.cc_max_flavor = self.config.get('cc_max_flavor')
+        # TODO validate early the existence of the max flavor
+        # what if it disappears afterwards?
+
+        self.deployment_repo_timeout = self.config.get(
+            'cc_deployment_repo_timeout', 20)
+        self.deployment_repo_wait_step = self.config.get(
+            'cc_deployment_repo_wait_step', 2)
 
     def init_cc_orga(self):
         if self.config.get('cc_multi_tenant'):
@@ -85,8 +94,7 @@ class CleverCloudDockerRunner(PaasRunner):
         """Cached dict of available flavors.
 
         This is a mapping whose keys are user flavor names, such as `'XL'`
-        and values are the actual names expected by Clever Cloud
-        (e.g, `'heptapod-runner-XL'`).
+        and values are :class:`CleverCloudFlavor` instances
 
         The caching is not really thread-safe, but the worst that can
         happen is several GET requests towards the Clever API if several
@@ -125,14 +133,21 @@ class CleverCloudDockerRunner(PaasRunner):
 
         prefix = self.config.get('cc_flavor_name_prefix', flavor_context + '-')
         available = {
-            flav['name'][len(prefix):]: flav['name']
+            flav['name'][len(prefix):]: CleverCloudFlavor(flav)
             for flav in resp.json()
-            if flav['name'].startswith(prefix)}
+            if flav['name'].startswith(prefix) and flav['available']}
         logger.info("Retrieved variant flavors. Resulting mapping: %r",
                     available)
         self._available_flavors = available, timestamp
         return available
 
+    @property
+    def min_requestable_weight(self):
+        if self.cc_max_flavor is not None:
+            return self.available_flavors[self.cc_max_flavor].weight
+
+        return max(fl.weight for fl in self.available_flavors.values())
+
     def init_cc_git(self):
         self.cc_git_user = self.config.get('cc_auth_user', 'Jenkins')
 
@@ -159,6 +174,11 @@ class CleverCloudDockerRunner(PaasRunner):
                 and not key.startswith('cc_'))
 
     def cc_docker_instance_type(self):
+        """Retrieve the CC instance type that we'll be using, with details.
+
+        The version is crucial, as we need to specify a supported one upon
+        provisioning. See also #37 about caching.
+        """
         resp = requests.get(self.cc_base_url + '/products/instances')
         if resp.status_code >= 400:
             error_code, message = convert_http_error(resp)
@@ -175,7 +195,7 @@ class CleverCloudDockerRunner(PaasRunner):
         # possible at a given time?
         return docker_types[0]
 
-    def cc_select_flavor(self, instance_type, job):
+    def cc_select_flavor(self, instance_type, job, check_max=True):
         flavor = get_job_variable(job, FLAVOR_JOB_VARIABLE_NAME)
         if flavor is None:
             flavor = self.cc_default_flavor
@@ -192,6 +212,22 @@ class CleverCloudDockerRunner(PaasRunner):
                                              list(self.available_flavors)),
                 executor=self.executor,
             )
+        if check_max and self.cc_max_flavor is not None:
+            max_flavor = self.available_flavors[self.cc_max_flavor]
+            if cc_flavor.weight > max_flavor.weight:
+                raise PaasProvisioningError(
+                    action='check-flavor',
+                    transport_code=None,
+                    code=2,
+                    error_details=f"Flavor size exceeded: this runner is "
+                    f"configured to have a maximum Clever Cloud flavor of "
+                    f"'{self.cc_max_flavor}', but this job specifies "
+                    f"the bigger flavor '{flavor}'. Please change the value "
+                    f"of the {FLAVOR_JOB_VARIABLE_NAME} job variable or "
+                    f"select another runner if possible.",
+                    executor=self.executor,
+                )
+
         return cc_flavor
 
     def job_cc_orga(self, job):
@@ -259,7 +295,7 @@ class CleverCloudDockerRunner(PaasRunner):
         req_data['instanceVersion'] = instance_type['version']
 
         flavor = self.cc_select_flavor(instance_type, job)
-        req_data['minFlavor'] = req_data['maxFlavor'] = flavor
+        req_data['minFlavor'] = req_data['maxFlavor'] = flavor.api_name
 
         req_data['name'] = 'hpd-job-%s-%d' % (self.unique_name, job['id'])
         app = self.job_cc_orga(job).create_app(req_data)
@@ -272,6 +308,15 @@ class CleverCloudDockerRunner(PaasRunner):
         app.put_env(cc_env)
         return app
 
+    @overrides
+    def expected_weight(self, job):
+        # TODO find a way to cache this call (with proper invalidation):
+        instance_type = self.cc_docker_instance_type()
+        # we'll let the launching thread check the max flavor and
+        # provide end-user feedback.
+        flavor = self.cc_select_flavor(instance_type, job, check_max=False)
+        return flavor.weight
+
     @overrides
     def load_paas_resource(self, data):
         # this maintains compatibility with data dumped without a
@@ -295,19 +340,67 @@ class CleverCloudDockerRunner(PaasRunner):
 
         return cc_orga.load_application(data)
 
+    def wait_deployability(self, app):
+        app_id = app.app_id
+        wait_step = self.deployment_repo_wait_step
+        timeout = self.deployment_repo_timeout
+
+        start = time.time()
+        while (time.time() - start) < timeout:
+            try:
+                state = app.get_info()['deployment']['repoState']
+            except KeyError:
+                raise PaasResourceError(
+                    app_id, self.executor,
+                    action='wait-deployability',
+                    error_details=f"Clever Cloud Application {app_id}: "
+                    "deployment details schema not understood",
+                    code=3,
+                    transport_code=None,
+                )
+            if state == 'CREATED':
+                break
+            time.sleep(wait_step)
+        else:
+            raise PaasResourceError(
+                app_id, self.executor,
+                action='wait-git',
+                error_details='Clever Cloud Application '
+                f'{app_id}: not deployable after {timeout} seconds',
+                code=4,
+                transport_code=None
+            )
+
     @overrides
     def launch(self, paas_resource, job_data):
+        self.wait_deployability(paas_resource)
+
         with tempfile.TemporaryDirectory() as tmp_path:
             build_helper = DockerBuildHelper.from_runner_config(tmp_path,
                                                                 self.config)
             build_helper.write_build_context(self, job_data)
-            build_helper.git_push(paas_resource.git_push_url)
+            try:
+                build_helper.git_push(paas_resource.git_push_url)
+            except DeploymentBranchAlreadyExisting:
+                logger.warning(
+                    "%s: depoyment branch already existing on remote. "
+                    "There probably was an issue with state loading "
+                    "or shutdown (SIGKILL or bug in shutdown sequence)",
+                    paas_resource.log_fmt()
+                )
+
+        paas_resource.launched = True
 
     @overrides
     def decommission(self, paas_resource):
         app_id = paas_resource.app_id
 
         resp = paas_resource.orga.delete_app(app_id)
+        if resp.status_code == 404:
+            logger.warning("decommision: application %r is already "
+                           "not present, nothing to be done", app_id)
+            return False
+
         if resp.status_code >= 400:
             error_code, message = convert_http_error(resp)
             raise PaasResourceError(app_id, self.executor,
@@ -317,6 +410,7 @@ class CleverCloudDockerRunner(PaasRunner):
                                     transport_code=resp.status_code,
                                     error_details=message)
         logger.debug('decommission: successful response %r', resp.text)
+        return True
 
     def decommission_all(self, gitlab_namespace=None):
         """Decommission all PAAS resources for this runner.
@@ -358,7 +452,9 @@ class CleverCloudDockerRunner(PaasRunner):
                 continue
 
             app = cc_orga.application(app_data)
+            logger.info("Decommissioning app name=%r", app_data['name'])
             self.decommission(app)
+            time.sleep(10)
             done += 1
 
         logger.info("Decommissionned the %d applications spawned by "
@@ -376,6 +472,19 @@ class CleverCloudOrganization:
     executor = CleverCloudDockerRunner.executor
     """Used in logging and raising of exceptions."""
 
+    transitional_default_weight = 32768
+    """Weight to attribute to applications created before weight tracking.
+
+    This is a worst-case value, meant to avoid over-provisionning due
+    to underestimation of currently running jobs at the first restart
+    with the weighted quota system.
+
+    We could do better than that, by querying actual flavors in service
+    upon state load, but it isn't worth the effort at this point: all we'll
+    have if there are lots of running jobs is a temporary saturation, until
+    at least one finishes.
+    """
+
     def __init__(self, base_api_url, orga_id, token, git_user,
                  gitlab_namespace=None):
         """Information needed to work on Clever API for an Organization.
@@ -438,6 +547,7 @@ class CleverCloudOrganization:
             deploy_url=resp_data['deployment']['httpUrl'],
             user=self.git_user,
             password=self.token,
+            weight=CleverCloudFlavor(resp_data['instance']['maxFlavor']).weight
         )
 
     def load_application(self, data):
@@ -447,14 +557,16 @@ class CleverCloudOrganization:
             app_id=data['id'],
             deploy_url=data['deploy_url'],
             user=self.git_user,
+            launched=data['launched'],
             password=self.token,
+            weight=data.get('weight', self.transitional_default_weight),
         )
 
     def delete_app(self, app_id):
         """Delete an application.
 
         Kept as a method on the organization, because we need less
-        detailed information to delete an application that to use it.
+        detailed information to delete an application than to use it.
         """
         return requests.delete(self.app_url(app_id),
                                headers=self.headers)
@@ -505,7 +617,9 @@ class OrganizationNotFound(CleverCloudOrganization):
 
 class CleverCloudApplication:
 
-    def __init__(self, orga, app_id, deploy_url, user, password):
+    def __init__(self, orga, app_id, deploy_url, user, password,
+                 launched=False,
+                 weight=None):
         self.orga = orga
         self.app_id = app_id
         self.api_url = orga.app_url(self.app_id)
@@ -514,6 +628,8 @@ class CleverCloudApplication:
         self.git_push_url = deploy_url.replace(
             'https://',
             'https://%s:%s@' % (user, password))
+        self.weight = weight
+        self.launched = launched
 
     @property
     def executor(self):
@@ -528,8 +644,36 @@ class CleverCloudApplication:
         # it will reinitialize the credentials, maybe with an updated value.
         return dict(id=self.app_id,
                     gitlab_namespace=self.orga.gitlab_namespace,
+                    weight=self.weight,
+                    launched=self.launched,
                     deploy_url=self.deploy_url)
 
+    def log_fmt(self):
+        """String representation useful for debug logs.
+
+        Not included attributes or fields:
+
+        - :attr:`deploy_url`: can be found by requesting the CC API if needed.
+        - :attr:`gitlab_namespace`: can be found by looking at the job in the
+          GitLab UI (or API)
+        """
+        return (f'CleverCloudApplication(app_id={self.app_id!r}, '
+                f'weight={self.weight:.2f}, '
+                f'launched={self.launched})')
+
+    def get_info(self):
+        resp = requests.get(self.api_url,
+                            headers=self.headers)
+        if resp.status_code >= 400:
+            error_code, message = convert_http_error(resp)
+            raise PaasResourceError(self.app_id, self.executor,
+                                    action='put-env',
+                                    action_details=None,
+                                    code=error_code,
+                                    transport_code=resp.status_code,
+                                    error_details=message)
+        return resp.json()
+
     def put_env(self, env):
         """Add environment variables to the application.
         """
@@ -549,3 +693,46 @@ class CleverCloudApplication:
         logger.debug('CleverCloudApplication.put_env: successful response %r',
                      resp.text)
         return resp
+
+
+class CleverCloudFlavor:
+    def __init__(self, api_data):
+        """Initialize from data returned by Clever Cloud API.
+
+        Could be from the flavor listing request, or from one of the flavors
+        listed in an application
+
+        Attributes:
+
+        - ``api_name``: the actual names expected by Clever Cloud API
+          (e.g, `'heptapod-runner-XL'`)
+        - ``ram``: allocated RAM, in MiB
+        - ``cpus``: number of virtual cores
+        """
+        # TODO actual response also includes a detailed `memory` key with
+        # values such as {'unit': 'B', value: '4294967296'}
+        # using it could be more robust than relying on the fact that `mem`
+        # is in MiB (we'd just need to implement all decimal and binary units
+        # (e.g., GB, kiB etc)
+        self.api_name = api_data['name']
+        self.ram_mib = api_data['mem']
+        self.cpus = api_data['cpus']
+
+    @property
+    def weight(self):
+        """Based on RAM footprint, with bias favoring small instances.
+
+        The reason for the bias is that it is typically harder to fit
+        large virtual machines in the available room on hypervisors. Of course
+        this merely helping, not providing any kind of certainty about the
+        provisionability.
+
+        The current formula makes it so that a flavor taking 16 times the
+        RAM of a smaller one counts 2 times as much as 16 of the smaller.
+        """
+        return (self.ram_mib / 1024) ** 1.25
+
+    def __repr__(self):
+        return "CleverCloudFlavor({'name': '%s', 'mem': %d, 'cpus': %d})" % (
+            self.api_name, self.ram_mib, self.cpus
+        )
